/*****************************************************************************
 * HASM64                                                                    *
 *****************************************************************************/

/*****************************************************************************
 * File name:     codegen.h                                                  *
 * Created:       2001-05-28 by Hampa Hug <hhug@student.ethz.ch>             *
 * Last modified: 2002-08-27 by Hampa Hug <hhug@student.ethz.ch>             *
 * Copyright:     (C) 2001-2002 by Hampa Hug <hhug@student.ethz.ch>          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by  the Free Software Foundation.                                         *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: codegen.h 10 2004-01-02 00:48:42Z hampa $ */


#ifndef HASM64_CODEGEN_H
#define HASM64_CODEGEN_H 1


typedef struct {
  unsigned      addr;
  unsigned      buf_i1;
  unsigned      buf_i2;
  unsigned long buf_n;
  unsigned char *buf;
} codegen_t;


/*!************************************************************************
 * Create a new code generator with the initial buffer size n.
 **************************************************************************/
codegen_t *cg_new (unsigned long n);

/*!************************************************************************
 * Delete a code generator.
 **************************************************************************/
void cg_del (codegen_t *cg);

/*!************************************************************************
 * Clear the output buffer.
 **************************************************************************/
void cg_clear (codegen_t *cg);

/*!************************************************************************
 * Set the current address. This does not affect the output buffer.
 **************************************************************************/
void cg_set_addr (codegen_t *cg, unsigned offs);

/*!************************************************************************
 * Get the current address.
 **************************************************************************/
unsigned cg_get_addr (codegen_t *cg);

/*!************************************************************************
 * Get the current index. This is the number of bytes written to the
 * output buffer so far.
 **************************************************************************/
unsigned cg_get_index (codegen_t *cg);

/*!************************************************************************
 * Add a byte cnt times to the output buffer.
 **************************************************************************/
int cg_add_byte (codegen_t *cg, unsigned char val, unsigned cnt);

/*!************************************************************************
 * Reserve cnt bytes in the output buffer.
 **************************************************************************/
int cg_res_byte (codegen_t *cg, unsigned cnt);

int cg_rep_bytes (codegen_t *cg, unsigned i1, unsigned i2, unsigned cnt);

/*!************************************************************************
 * Write the output buffer to a file.
 **************************************************************************/
int cg_write (codegen_t *cg, FILE *fp);

/*!************************************************************************
 * Write the output buffer to a file.
 **************************************************************************/
int cg_write_file (codegen_t *cg, char *fname);


#endif
