/*****************************************************************************
 * HASM64                                                                    *
 *****************************************************************************/

/*****************************************************************************
 * File name:     hdasm64.c                                                  *
 * Created:       2001-06-01 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2004-01-02 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2001-2004 by Hampa Hug <hampa@hampa.ch>                *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by  the Free Software Foundation.                                         *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: hdasm64.c 12 2004-01-02 02:11:34Z hampa $ */


#include <stdio.h>

#include "hdasm64.h"


static int      h64_verbose = 0;
static unsigned h64_base = 0;
static char     *h64_mod_name = NULL;


void prt_help (void)
{
  fprintf (stderr,
    "Usage: hdasm64 options\n\n"
    "Options are:\n"
    "  -i string            Set input file name [stdin]\n"
    "  -o string            Set output file name [stdout]\n"
    "  -b, --base int       Set base address [0]\n"
    "  -m, --module string  Set module name [noname]\n"
  );
  fflush (stderr);
}

static
void prt_version (void)
{
  fputs (
    "hdasm64 version " HASM64_VERSION_STR
    " (" HASM64_CFG_DATE " " HASM64_CFG_TIME ")\n\n"
    "Copyright (C) 2001-2004 by Hampa Hug <hampa@hampa.ch>\n",
    stdout
  );

  fflush (stdout);
}

void prt_error (char *msg)
{
  fprintf (stderr, "hdasm64: %s\n", msg);
  fflush (stderr);
}

char *str_clone (char *str)
{
  char *ret;

  ret = (char *) malloc (strlen (str) + 1);
  if (ret == NULL) {
    return (NULL);
  }

  strcpy (ret, str);

  return (ret);
}

int str_isarg1 (char *str, char *arg)
{
  if (strcmp (str, arg) == 0) {
    return (1);
  }

  return (0);
}

int str_isarg2 (char *str, char *arg1, char *arg2)
{
  if (strcmp (str, arg1) == 0) {
    return (1);
  }

  if (strcmp (str, arg2) == 0) {
    return (1);
  }

  return (0);
}


static op_tab_t op_tab[256] = {
  { 0x00, "brk", 1, M_IMP },
  { 0x01, "ora", 2, M_IND_X },
  { 0x02, NULL,  1, M_IMP },
  { 0x03, NULL,  1, M_IMP },
  { 0x04, NULL,  1, M_IMP },
  { 0x05, "ora", 2, M_ZPG },
  { 0x06, "asl", 2, M_ZPG },
  { 0x07, NULL,  1, M_IMP },
  { 0x08, "php", 1, M_IMP },
  { 0x09, "ora", 2, M_IMM },
  { 0x0a, "asl", 1, M_ACC },
  { 0x0b, NULL,  1, M_IMP },
  { 0x0c, NULL,  1, M_IMP },
  { 0x0d, "ora", 3, M_ABS },
  { 0x0e, "asl", 3, M_ABS },
  { 0x0f, NULL,  1, M_IMP },
  { 0x10, "bpl", 2, M_REL },
  { 0x11, "ora", 2, M_IND_Y },
  { 0x12, NULL,  1, M_IMP },
  { 0x13, NULL,  1, M_IMP },
  { 0x14, NULL,  1, M_IMP },
  { 0x15, "ora", 2, M_ZPG_X },
  { 0x16, "asl", 2, M_ZPG_X },
  { 0x17, NULL,  1, M_IMP },
  { 0x18, "clc", 1, M_IMP },
  { 0x19, "ora", 3, M_ABS_Y },
  { 0x1a, NULL,  1, M_IMP },
  { 0x1b, NULL,  1, M_IMP },
  { 0x1c, NULL,  1, M_IMP },
  { 0x1d, "ora", 3, M_ABS_X },
  { 0x1e, "asl", 3, M_ABS_X },
  { 0x1f, NULL,  1, M_IMP },
  { 0x20, "jsr", 3, M_IMM16 },
  { 0x21, "and", 2, M_IND_X },
  { 0x22, NULL,  1, M_IMP },
  { 0x23, NULL,  1, M_IMP },
  { 0x24, "bit", 2, M_ZPG },
  { 0x25, "and", 2, M_ZPG },
  { 0x26, "rol", 2, M_ZPG },
  { 0x27, NULL,  1, M_IMP },
  { 0x28, "plp", 1, M_IMP },
  { 0x29, "and", 2, M_IMM },
  { 0x2a, "rol", 1, M_ACC },
  { 0x2b, NULL,  1, M_IMP },
  { 0x2c, "bit", 3, M_ABS },
  { 0x2d, "and", 3, M_ABS },
  { 0x2e, "rol", 3, M_ABS },
  { 0x2f, NULL,  1, M_IMP },
  { 0x30, "bmi", 2, M_REL },
  { 0x31, "and", 2, M_IND_Y },
  { 0x32, NULL,  1, M_IMP },
  { 0x33, NULL,  1, M_IMP },
  { 0x34, NULL,  1, M_IMP },
  { 0x35, "and", 2, M_ZPG_X },
  { 0x36, "rol", 2, M_ZPG_X },
  { 0x37, NULL,  1, M_IMP },
  { 0x38, "sec", 1, M_IMP },
  { 0x39, "and", 3, M_ABS_Y },
  { 0x3a, NULL,  1, M_IMP },
  { 0x3b, NULL,  1, M_IMP },
  { 0x3c, NULL,  1, M_IMP },
  { 0x3d, "and", 3, M_ABS_X },
  { 0x3e, "rol", 3, M_ABS_X },
  { 0x3f, NULL,  1, M_IMP },
  { 0x40, "rti", 1, M_IMP },
  { 0x41, "eor", 2, M_IND_X },
  { 0x42, NULL,  1, M_IMP },
  { 0x43, NULL,  1, M_IMP },
  { 0x44, NULL,  1, M_IMP },
  { 0x45, "eor", 2, M_ZPG },
  { 0x46, "lsr", 2, M_ZPG },
  { 0x47, NULL,  1, M_IMP },
  { 0x48, "pha", 1, M_IMP },
  { 0x49, "eor", 2, M_IMM },
  { 0x4a, "lsr", 1, M_ACC },
  { 0x4b, NULL,  1, M_IMP },
  { 0x4c, "jmp", 3, M_IMM16 },
  { 0x4d, "eor", 3, M_ABS },
  { 0x4e, "lsr", 3, M_ABS },
  { 0x4f, NULL,  1, M_IMP },
  { 0x50, "bvc", 2, M_REL },
  { 0x51, "eor", 2, M_IND_Y },
  { 0x52, NULL,  1, M_IMP },
  { 0x53, NULL,  1, M_IMP },
  { 0x54, NULL,  1, M_IMP },
  { 0x55, "eor", 2, M_ZPG_X },
  { 0x56, "lsr", 2, M_ZPG_X },
  { 0x57, NULL,  1, M_IMP },
  { 0x58, "cli", 1, M_IMP },
  { 0x59, "eor", 3, M_ABS_Y },
  { 0x5a, NULL,  1, M_IMP },
  { 0x5b, NULL,  1, M_IMP },
  { 0x5c, NULL,  1, M_IMP },
  { 0x5d, "eor", 3, M_ABS_X },
  { 0x5e, "lsr", 3, M_ABS_X },
  { 0x5f, NULL,  1, M_IMP },
  { 0x60, "rts", 1, M_IMP },
  { 0x61, "adc", 2, M_IND_X },
  { 0x62, NULL,  1, M_IMP },
  { 0x63, NULL,  1, M_IMP },
  { 0x64, NULL,  1, M_IMP },
  { 0x65, "adc", 2, M_ZPG },
  { 0x66, "ror", 2, M_ZPG },
  { 0x67, NULL,  1, M_IMP },
  { 0x68, "pla", 1, M_IMP },
  { 0x69, "adc", 2, M_IMM },
  { 0x6a, "ror", 1, M_ACC },
  { 0x6b, NULL,  1, M_IMP },
  { 0x6c, "jmp", 3, M_ABS },
  { 0x6d, "adc", 3, M_ABS },
  { 0x6e, "ror", 3, M_ABS },
  { 0x6f, NULL,  1, M_IMP },
  { 0x70, "bvs", 2, M_REL },
  { 0x71, "adc", 2, M_IND_Y },
  { 0x72, NULL,  1, M_IMP },
  { 0x73, NULL,  1, M_IMP },
  { 0x74, NULL,  1, M_IMP },
  { 0x75, "adc", 2, M_ZPG_X },
  { 0x76, "ror", 2, M_ZPG_X },
  { 0x77, NULL,  1, M_IMP },
  { 0x78, "sei", 1, M_IMP },
  { 0x79, "adc", 3, M_ABS_Y },
  { 0x7a, NULL,  1, M_IMP },
  { 0x7b, NULL,  1, M_IMP },
  { 0x7c, NULL,  1, M_IMP },
  { 0x7d, "adc", 3, M_ABS_X },
  { 0x7e, "ror", 3, M_ABS_X },
  { 0x7f, NULL,  1, M_IMP },
  { 0x80, NULL,  1, M_IMP },
  { 0x81, "sta", 2, M_IND_X },
  { 0x82, NULL,  1, M_IMP },
  { 0x83, NULL,  1, M_IMP },
  { 0x84, "sty", 2, M_ZPG },
  { 0x85, "sta", 2, M_ZPG },
  { 0x86, "stx", 2, M_ZPG },
  { 0x87, NULL,  1, M_IMP },
  { 0x88, "dey", 1, M_IMP },
  { 0x89, NULL,  1, M_IMP },
  { 0x8a, "txa", 1, M_IMP },
  { 0x8b, NULL,  1, M_IMP },
  { 0x8c, "sty", 3, M_ABS },
  { 0x8d, "sta", 3, M_ABS },
  { 0x8e, "stx", 3, M_ABS },
  { 0x8f, NULL,  1, M_IMP },
  { 0x90, "bcc", 2, M_REL },
  { 0x91, "sta", 2, M_IND_Y },
  { 0x92, NULL,  1, M_IMP },
  { 0x93, NULL,  1, M_IMP },
  { 0x94, "sty", 2, M_ZPG_X },
  { 0x95, "sta", 2, M_ZPG_X },
  { 0x96, "stx", 2, M_ZPG_Y },
  { 0x97, NULL,  1, M_IMP },
  { 0x98, "tya", 1, M_IMP },
  { 0x99, "sta", 3, M_ABS_Y },
  { 0x9a, "txs", 1, M_IMP },
  { 0x9b, NULL,  1, M_IMP },
  { 0x9c, NULL,  1, M_IMP },
  { 0x9d, "sta", 3, M_ABS_X },
  { 0x9e, NULL,  1, M_IMP },
  { 0x9f, NULL,  1, M_IMP },
  { 0xa0, "ldy", 2, M_IMM },
  { 0xa1, "lda", 2, M_IND_X },
  { 0xa2, "ldx", 2, M_IMM },
  { 0xa3, NULL,  1, M_IMP },
  { 0xa4, "ldy", 2, M_ZPG },
  { 0xa5, "lda", 2, M_ZPG },
  { 0xa6, "ldx", 2, M_ZPG },
  { 0xa7, NULL,  1, M_IMP },
  { 0xa8, "tay", 1, M_IMP },
  { 0xa9, "lda", 2, M_IMM },
  { 0xaa, "tax", 1, M_IMP },
  { 0xab, NULL,  1, M_IMP },
  { 0xac, "ldy", 3, M_ABS },
  { 0xad, "lda", 3, M_ABS },
  { 0xae, "ldx", 3, M_ABS },
  { 0xaf, NULL,  1, M_IMP },
  { 0xb0, "bcs", 2, M_REL },
  { 0xb1, "lda", 2, M_IND_Y },
  { 0xb2, NULL,  1, M_IMP },
  { 0xb3, NULL,  1, M_IMP },
  { 0xb4, "ldy", 2, M_ZPG_X },
  { 0xb5, "lda", 2, M_ZPG_X },
  { 0xb6, "ldx", 2, M_ZPG_Y },
  { 0xb7, NULL,  1, M_IMP },
  { 0xb8, "clv", 1, M_IMP },
  { 0xb9, "lda", 3, M_ABS_Y },
  { 0xba, "tsx", 1, M_IMP },
  { 0xbb, NULL,  1, M_IMP },
  { 0xbc, "ldy", 3, M_ABS_X },
  { 0xbd, "lda", 3, M_ABS_X },
  { 0xbe, "ldx", 3, M_ABS_Y },
  { 0xbf, NULL,  1, M_IMP },
  { 0xc0, "cpy", 2, M_IMM },
  { 0xc1, "cmp", 2, M_IND_X },
  { 0xc2, NULL,  1, M_IMP },
  { 0xc3, NULL,  1, M_IMP },
  { 0xc4, "cpy", 2, M_ZPG },
  { 0xc5, "cmp", 2, M_ZPG },
  { 0xc6, "dec", 2, M_ZPG },
  { 0xc7, NULL,  1, M_IMP },
  { 0xc8, "iny", 1, M_IMP },
  { 0xc9, "cmp", 2, M_IMM },
  { 0xca, "dex", 1, M_IMP },
  { 0xcb, NULL,  1, M_IMP },
  { 0xcc, "cpy", 3, M_ABS },
  { 0xcd, "cmp", 3, M_ABS },
  { 0xce, "dec", 3, M_ABS },
  { 0xcf, NULL,  1, M_IMP },
  { 0xd0, "bne", 2, M_REL },
  { 0xd1, "cmp", 2, M_IND_Y },
  { 0xd2, NULL,  1, M_IMP },
  { 0xd3, NULL,  1, M_IMP },
  { 0xd4, NULL,  1, M_IMP },
  { 0xd5, "cmp", 2, M_ZPG_X },
  { 0xd6, "dec", 2, M_ZPG_X },
  { 0xd7, NULL,  1, M_IMP },
  { 0xd8, "cld", 1, M_IMP },
  { 0xd9, "cmp", 3, M_ABS_Y },
  { 0xda, NULL,  1, M_IMP },
  { 0xdb, NULL,  1, M_IMP },
  { 0xdc, NULL,  1, M_IMP },
  { 0xdd, "cmp", 3, M_ABS_X },
  { 0xde, "dec", 3, M_ABS_X },
  { 0xdf, NULL,  1, M_IMP },
  { 0xe0, "cpx", 2, M_IMM },
  { 0xe1, "sbc", 2, M_IND_X },
  { 0xe2, NULL,  1, M_IMP },
  { 0xe3, NULL,  1, M_IMP },
  { 0xe4, "cpx", 2, M_ZPG },
  { 0xe5, "sbc", 2, M_ZPG },
  { 0xe6, "inc", 2, M_ZPG },
  { 0xe7, NULL,  1, M_IMP },
  { 0xe8, "inx", 1, M_IMP },
  { 0xe9, "sbc", 2, M_IMM },
  { 0xea, "nop", 1, M_IMP },
  { 0xeb, NULL,  1, M_IMP },
  { 0xec, "cpx", 3, M_ABS },
  { 0xed, "sbc", 3, M_ABS },
  { 0xee, "inc", 3, M_ABS },
  { 0xef, NULL,  1, M_IMP },
  { 0xf0, "beq", 2, M_REL },
  { 0xf1, "sbc", 2, M_IND_Y },
  { 0xf2, NULL,  1, M_IMP },
  { 0xf3, NULL,  1, M_IMP },
  { 0xf4, NULL,  1, M_IMP },
  { 0xf5, "sbc", 2, M_ZPG_X },
  { 0xf6, "inc", 2, M_ZPG_X },
  { 0xf7, NULL,  1, M_IMP },
  { 0xf8, "sed", 1, M_IMP },
  { 0xf9, "sbc", 3, M_ABS_Y },
  { 0xfa, NULL,  1, M_IMP },
  { 0xfb, NULL,  1, M_IMP },
  { 0xfc, NULL,  1, M_IMP },
  { 0xfd, "sbc", 3, M_ABS_X },
  { 0xfe, "inc", 3, M_ABS_X },
  { 0xff, NULL,  1, M_IMP },
};

unsigned dis_op_xx (instr_t *op, unsigned addr, unsigned char *buf, unsigned n)
{
  op->addr = addr;
  op->cnt = 1;
  op->data[0] = buf[0];

  strcpy (op->op, "db");
  sprintf (op->par, "0x%02x", buf[0]);

  return (1);
}

unsigned dis_op (instr_t *op, unsigned addr, unsigned char *buf, unsigned n)
{
  unsigned i;
  unsigned cnt;
  unsigned op_i;
  unsigned tmp;

  op_i = buf[0];

  cnt = op_tab[op_i].cnt;

  if ((n < cnt) || (op_tab[op_i].instr == NULL)) {
    return (dis_op_xx (op, addr, buf, n));
  }

  op->addr = addr;
  op->cnt = cnt;
  for (i = 0; i < cnt; i++) {
    op->data[i] = buf[i];
  }

  strcpy (op->op, op_tab[op_i].instr);

  switch (op_tab[op_i].mode) {
    case M_IMP:
      strcpy (op->par, "");
      break;

    case M_ACC:
      sprintf (op->par, "a");
      break;

    case M_IMM:
      sprintf (op->par, "0x%02x", buf[1]);
      break;

    case M_ZPG:
      sprintf (op->par, "[0x%02x]", buf[1]);
      break;

    case M_ABS:
      if (buf[2] == 0) {
        strcpy (op->op, "db");
        sprintf (op->par, "0x%02x, 0x%02x, 0x%02x", buf[0], buf[1], buf[2]);
      }
      else {
        sprintf (op->par, "[0x%04x]", (buf[2] << 8) + buf[1]);
      }
      break;

    case M_ABS_X:
      if (buf[2] == 0) {
        strcpy (op->op, "db");
        sprintf (op->par, "0x%02x, 0x%02x, 0x%02x", buf[0], buf[1], buf[2]);
      }
      else {
        sprintf (op->par, "[0x%04x + x]", (buf[2] << 8) + buf[1]);
      }
      break;

    case M_ABS_Y:
      if (buf[2] == 0) {
        strcpy (op->op, "db");
        sprintf (op->par, "0x%02x, 0x%02x, 0x%02x", buf[0], buf[1], buf[2]);
      }
      else {
        sprintf (op->par, "[0x%04x + y]", (buf[2] << 8) + buf[1]);
      }
      break;

    case M_ZPG_X:
      sprintf (op->par, "[0x%02x + x]", buf[1]);
      break;

    case M_ZPG_Y:
      sprintf (op->par, "[0x%02x + y]", buf[1]);
      break;

    case M_IND_X:
      sprintf (op->par, "[[0x%02x + x]]", buf[1]);
      break;

    case M_IND_Y:
      sprintf (op->par, "[[0x%02x] + y]", buf[1]);
      break;

    case M_REL:
      if (buf[1] < 0x80) {
        tmp = buf[1];
      }
      else {
        tmp = 0xff00 | buf[1];
      }

      tmp = (addr + tmp + 2) & 0xffff;

      sprintf (op->par, "0x%04x", tmp);
      break;

    case M_IMM16:
      sprintf (op->par, "0x%04x", (buf[2] << 8) + buf[1]);
      break;

    default:
      break;
  }

  return (cnt);
}

void prt_instr (instr_t *op, FILE *f_out)
{
  unsigned i;

  fprintf (f_out, "  %-6s %-29s ; %04x ", op->op, op->par, op->addr);

  for (i = 0; i < op->cnt; i++) {
    fprintf (f_out, " %02x", op->data[i]);
  }

  fprintf (f_out, "\n");
}

int h64_disasm (FILE *f_inp, FILE *f_out)
{
  unsigned      i;
  unsigned      n, r;
  unsigned      addr;
  unsigned char buf[4];
  instr_t       op;

  addr = h64_base;

  fprintf (f_out, "module %s = 0x%04x\n", h64_mod_name, addr);

  n = fread (buf, 1, 4, f_inp);
  while (n > 0) {
    r = dis_op (&op, addr, buf, n);

    prt_instr (&op, f_out);

    for (i = r; i < 4; i++) {
      buf[i - r] = buf[i];
    }

    addr += r;
    n -= r;

    n += fread (buf + n, 1, 4 - n, f_inp);
  }

  fprintf (f_out, "end\n");

  return (0);
}

int main (int argc, char **argv)
{
  int       i, r;
  FILE      *f_inp[2];
  FILE      *f_out[2];

  if (argc == 2) {
    if (str_isarg1 (argv[1], "--help")) {
      prt_help ();
      return (0);
    }
    else if (str_isarg1 (argv[1], "--version")) {
      prt_version ();
      return (0);
    }
  }

  f_inp[0] = stdin;
  f_inp[1] = NULL;
  f_out[0] = stdout;
  f_out[1] = NULL;

  i = 1;
  while (i < argc) {
    if (str_isarg2 (argv[i], "-v", "--verbose")) {
      h64_verbose = 1;
    }
    else if (str_isarg1 (argv[i], "-i")) {
      i += 1;
      if (i >= argc) {
        prt_error ("Missing input file name");
        return (1);
      }

      f_inp[1] = fopen (argv[i], "rb");
      if (f_inp[1] == NULL) {
        prt_error ("Can't open input file");
        return (1);
      }

      f_inp[0] = f_inp[1];
    }
    else if (str_isarg1 (argv[i], "-o")) {
      i += 1;
      if (i >= argc) {
        prt_error ("Missing output file name");
        return (1);
      }

      f_out[1] = fopen (argv[i], "wb");
      if (f_out[1] == NULL) {
        prt_error ("Can't open output file");
        return (1);
      }

      f_out[0] = f_out[1];
    }
    else if (str_isarg2 (argv[i], "-b", "--base")) {
      i += 1;
      if (i >= argc) {
        prt_error ("Missing base address");
        return (1);
      }

      h64_base = (unsigned)strtoul (argv[i], NULL, 0);
    }
    else if (str_isarg2 (argv[i], "-m", "--module")) {
      i += 1;
      if (i >= argc) {
        prt_error ("Missing module name");
        return (1);
      }

      h64_mod_name = str_clone (argv[i]);
    }
    else {
      prt_error ("Unknown option");
      return (1);
    }

    i += 1;
  }

  if (h64_mod_name == NULL) {
    h64_mod_name = str_clone ("noname");
  }

  r = h64_disasm (f_inp[0], f_out[0]);

  if (f_out[1] != NULL) {
    fclose (f_out[1]);
  }

  if (f_inp[1] != NULL) {
    fclose (f_inp[1]);
  }

  free (h64_mod_name);

  return (r);
}
